@@ -3,10 +3,10 @@
  *		variously munged/massaged/relayered/slashed/burned
  *			since then. -pgf
  *
- *	term.getch()	raw 8-bit key from terminal driver.
+ *	term.getc()	raw 8-bit key from terminal driver.
  *
  *	sysmapped_c()	single "keystroke" -- may have SPEC bit, if it was
- *			a sytem-mapped function key.  calls term.getch().  these
+ *			a sytem-mapped function key.  calls term.getc().  these
  *			system-mapped keys will never map to a multi-char
  *			sequence.  the routine does have storage, to hold
  *			keystrokes gathered "in error".
@@ -39,12 +39,12 @@
  *			SPEC|c.
  *
  *
- *	term.typahead()	  true if a key is avail from term.getch().
+ *	term.typahead()	  true if a key is avail from term.getc().
  *	sysmapped_c_avail() "  if a key is avail from sysmapped_c() or below.
  *	tgetc_avail()     true if a key is avail from tgetc() or below.
  *	keystroke_avail() true if a key is avail from keystroke() or below.
  *
- * $Header: /usr/build/vile/vile/RCS/input.c,v 1.283 2005/05/27 01:44:55 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/input.c,v 1.316 2007/12/27 19:18:51 tom Exp $
  *
  */
 
@@ -122,9 +122,9 @@ int
 shell_complete(DONE_ARGS)
 {
     int status = FALSE;
-    unsigned len = *pos;
-    int base;
-    int first = 0;
+    size_t len = *pos;
+    size_t base;
+    size_t first = 0;
 
     TRACE(("shell_complete %d:'%s'\n", *pos, buf));
     if (buf != 0) {
@@ -550,7 +550,7 @@ tgetc(int quoted)
 	     * getchar() returns will be the same as the last
 	     * character before the interrupt.  Eat it.
 	     */
-	    (void) ttgetc();
+	    (void) term.getc();
 #endif
 	} else {
 	    (void) im_waiting(TRUE);
@@ -619,60 +619,185 @@ kbd_seq_nomap(void)
     return kbd_seq();
 }
 
-/* get a string consisting of inclchartype characters from the current
-	position.  if inclchartype is 0, return everything to eol */
+/*
+ * Check for special cases which override character-classes when getting
+ * data from the screen.
+ */
 int
-screen_string(char *buf, int bufn, CHARTYPE inclchartype)
+adjust_chartype(CHARTYPE *mask)
+{
+    int whole_line = FALSE;
+
+    /* if from gototag(), grab from the beginning of the string */
+    if (b_val(curbp, MDTAGWORD)
+	&& *mask == vl_ident) {
+	whole_line = TRUE;
+    } else if (b_is_directory(curbp)
+	       && *mask == SCREEN_STRING) {
+	whole_line = TRUE;
+	*mask = (CHARTYPE) ~0;
+    }
+    return whole_line;
+}
+
+/*
+ * Get a string consisting of inclchartype characters starting from the current
+ * position.  If inclchartype is 0, return everything to end of line.  If
+ * whole_line is true, allow the string to start before the current position.
+ */
+static int
+read_by_ctype(char *buf, size_t bufn, CHARTYPE inclchartype, int whole_line)
 {
     static TBUFF *temp;
     int rc;
-    int len;
+    size_t len;
 
-    if ((rc = screen2tbuff(&temp, inclchartype)) == TRUE) {
+#if OPT_CURTOKENS
+    if (b_val(curbp, VAL_CURSOR_TOKENS) == CT_REGEX) {
+	if (!(b_is_directory(curbp) && whole_line))
+	    return FALSE;
+    }
+#endif
+
+    TRACE((T_CALLED "read_by_ctype(incl=%#lx)\n", (ULONG) inclchartype));
+    if ((rc = vl_ctype2tbuff(&temp, inclchartype, whole_line)) == TRUE) {
 	if ((len = tb_length(temp)) >= bufn)
 	    len = bufn - 1;
 	strncpy0(buf, tb_values(temp), len + 1);
     } else {
 	*buf = EOS;
     }
-    return rc;
+    TRACE(("result %s\n", buf));
+    returnCode(rc);
 }
 
-/* get a string consisting of inclchartype characters from the current
-	position.  if inclchartype is 0, return everything to eol */
+#if OPT_CURTOKENS
+static int
+read_by_regex(char *buf, size_t bufn, REGEXVAL * rexp, int whole_line)
+{
+    static TBUFF *temp;
+    int rc = FALSE;
+    size_t len;
+
+    if (b_val(curbp, VAL_CURSOR_TOKENS) != CT_CCLASS && rexp != 0) {
+	TRACE((T_CALLED "read_by_regex(incl=%s)\n", rexp->pat));
+	*buf = EOS;
+
+	if (*rexp->pat != EOS && rexp->reg != 0) {
+	    if ((rc = vl_regex2tbuff(&temp, rexp, whole_line)) == TRUE) {
+		if ((len = tb_length(temp)) >= bufn)
+		    len = bufn - 1;
+		strncpy0(buf, tb_values(temp), len + 1);
+	    }
+	}
+	TRACE(("result %s\n", buf));
+    }
+    returnCode(rc);
+}
+#else
+#define read_by_regex(buf, bufn, rexp, whole_line) FALSE
+#endif
+
+/*
+ * Obtain a buffer (or file) name from the screen.
+ * It may contain wildcards that we can glob-expand.
+ *
+ * If it is not a directory-buffer, we provide a choice between regular
+ * expressions and character classes.  Directory-buffers are guaranteed
+ * to have one filename per line.
+ */
 int
-screen2tbuff(TBUFF **result, CHARTYPE inclchartype)
+screen_to_bname(char *buf, size_t bufn)
+{
+    int rc;
+    CHARTYPE mask = SCREEN_STRING;
+    int whole_line = adjust_chartype(&mask);
+
+    TRACE((T_CALLED "screen_to_bname\n"));
+#if OPT_CURTOKENS
+    if (b_is_directory(curbp))
+	rc = FALSE;
+    else
+	rc = read_by_regex(buf, bufn, get_buf_fname_expr(curbp), whole_line);
+    if (rc == FALSE)
+#endif
+	rc = read_by_ctype(buf, bufn, mask, whole_line);
+    returnCode(rc);
+}
+
+/*
+ * Obtain $pathname from the screen.
+ */
+int
+screen_to_fname(char *buf, size_t bufn)
+{
+    int rc = FALSE;
+    CHARTYPE mask = vl_pathn;
+    int whole_line = adjust_chartype(&mask);
+
+    TRACE((T_CALLED "screen_to_fname\n"));
+    rc = read_by_regex(buf, bufn, b_val_rexp(curbp, VAL_PATHNAME_EXPR), whole_line);
+    if (rc == FALSE)
+	rc = read_by_ctype(buf, bufn, mask, whole_line);
+
+    returnCode(rc);
+}
+
+/*
+ * Obtain $identifier from the screen.
+ */
+int
+screen_to_ident(char *buf, size_t bufn)
+{
+    int rc = FALSE;
+    CHARTYPE mask = vl_ident;
+    int whole_line = adjust_chartype(&mask);
+
+    TRACE((T_CALLED "screen_to_ident\n"));
+    rc = read_by_regex(buf, bufn, b_val_rexp(curbp, VAL_IDENTIFIER_EXPR), whole_line);
+    if (rc == FALSE)
+	rc = read_by_ctype(buf, bufn, mask, whole_line);
+    returnCode(rc);
+}
+
+/*
+ * Get a string consisting of inclchartype characters starting from the current
+ * position.  If inclchartype is 0, return everything to end of line.  If
+ * whole_line is true, allow the string to start before the current position.
+ */
+int
+vl_ctype2tbuff(TBUFF **result, CHARTYPE inclchartype, int whole_line)
 {
     int i = 0;
     MARK save_dot;
     int first = -1;
     int last = -1;
-    int whole_line = 0;
+
+    TRACE((T_CALLED "vl_ctype2tbuff(incl=%#lx)\n", (ULONG) inclchartype));
 
     save_dot = DOT;
 
     tb_init(result, EOS);
-    /* if from gototag(), grab from the beginning of the string */
-    if (b_val(curbp, MDTAGWORD)
-	&& inclchartype == vl_ident) {
-	whole_line = 1;
-    } else if (b_is_directory(curbp)
-	       && inclchartype == SCREEN_STRING) {
-	whole_line = 1;
-	inclchartype = (CHARTYPE) ~0;
-    }
 
+    /*
+     * If we're processing the whole line, find the beginning of the token
+     * by searching to the left.
+     */
     if (whole_line
 	&& DOT.o > 0
 	&& istype(inclchartype, char_at(DOT))) {
 	while (DOT.o > 0) {
-	    DOT.o--;
+	    backchar_to_bol(TRUE, 1);
 	    if (!istype(inclchartype, char_at(DOT))) {
-		DOT.o++;
+		forwchar_to_eol(TRUE, 1);
 		break;
 	    }
 	}
     }
+
+    /*
+     * Search for the start of the token, and then the end of the token.
+     */
     while (!is_at_end_of_line(DOT)) {
 	last = CharOf(char_at(DOT));
 	if (first < 0)
@@ -704,7 +829,8 @@ screen2tbuff(TBUFF **result, CHARTYPE inclchartype)
 	if (inclchartype && !istype(inclchartype, last))
 	    break;
 	tb_append(result, last);
-	DOT.o++;
+	if (!forwchar_to_eol(TRUE, 1))
+	    break;
 	i++;
 #if OPT_WIDE_CTYPES
 	if (inclchartype & vl_scrtch) {
@@ -732,9 +858,55 @@ screen2tbuff(TBUFF **result, CHARTYPE inclchartype)
 
     DOT = save_dot;
 
-    return tb_length(*result) != 0;
+    returnCode(tb_length(*result) != 0);
 }
 
+/*
+ * Get a string matching the given expression, which includes the current
+ * position.  If the "whole_line" parameter is true, allow the string to
+ * start before the current position.
+ */
+#if OPT_CURTOKENS
+int
+vl_regex2tbuff(TBUFF **result, REGEXVAL * rexp, int whole_line)
+{
+    C_NUM given = DOT.o;
+    C_NUM length;
+    C_NUM offset;
+    C_NUM match_off = -1;
+    C_NUM match_len = -1;
+
+    TRACE((T_CALLED "vl_regex2tbuff\n"));
+
+    tb_init(result, EOS);
+
+    if (rexp != 0 && rexp->pat != 0 && rexp->pat[0] && rexp->reg) {
+	regexp *exp = rexp->reg;
+	char *line_text = lvalue(DOT.l);
+
+	while (given >= 0) {
+	    if (lregexec(exp, DOT.l, given, llength(DOT.l))) {
+		offset = exp->startp[0] - line_text;
+		length = exp->endp[0] - exp->startp[0];
+		if ((length > match_len) && (offset + length >= DOT.o)) {
+		    match_off = offset;
+		    match_len = length;
+		}
+	    }
+	    if (!whole_line)
+		break;
+	    --given;
+	}
+	if (match_len > 0) {
+	    tb_bappend(result, line_text + match_off, match_len);
+	    tb_append(result, EOS);
+	}
+    }
+
+    returnCode(tb_length(*result) != 0);
+}
+#endif /* OPT_CURTOKENS */
+
 /*
  * Returns the character that ended the last call on 'kbd_string()'
  */
@@ -870,11 +1042,11 @@ remove_backslashes(TBUFF *buf)
 /* count backslashes so we can tell at any point whether we have the current
  * position escaped by one.
  */
-static UINT
-countBackSlashes(TBUFF *buf, UINT len)
+static size_t
+countBackSlashes(TBUFF *buf, size_t len)
 {
     char *buffer = tb_values(buf);
-    UINT count;
+    size_t count;
 
     if (len && buffer != 0 && isEscaped(buffer + len)) {
 	count = 1;
@@ -924,11 +1096,11 @@ show1Char(int c)
 /* expand a single character (only used on interactive input) */
 static int
 expandChar(TBUFF **buf,
-	   unsigned *position,
+	   size_t *position,
 	   int c,
 	   KBD_OPTIONS options)
 {
-    int cpos = *position;
+    int cpos = (int) *position;
     char *cp;
     BUFFER *bp;
     char str[NFILEN];
@@ -978,7 +1150,7 @@ expandChar(TBUFF **buf,
 	if (!expand)
 	    return FALSE;
 
-	if (screen_string(str, sizeof(str), vl_pathn))
+	if (screen_to_fname(str, sizeof(str)))
 	    cp = str;
 	else
 	    cp = NULL;
@@ -1042,11 +1214,11 @@ is_edit_char(int c)
  * Erases the response from the screen for 'kbd_string()'
  */
 void
-kbd_kill_response(TBUFF *buffer, unsigned *position, int c)
+kbd_kill_response(TBUFF *buffer, size_t *position, int c)
 {
     char *buf = tb_values(buffer);
     TBUFF *tmp = 0;
-    int cpos = *position;
+    int cpos = (int) *position;
     UINT mark = cpos;
 
     if (buf != 0) {
@@ -1083,11 +1255,11 @@ kbd_kill_response(TBUFF *buffer, unsigned *position, int c)
 int
 kbd_show_response(TBUFF **dst,	/* string with escapes */
 		  char *src,	/* string w/o escapes */
-		  unsigned bufn,	/* # of chars we read from 'src[]' */
+		  size_t bufn,	/* # of chars we read from 'src[]' */
 		  int eolchar,
 		  KBD_OPTIONS options)
 {
-    unsigned k;
+    size_t k;
 
     /* add backslash escapes in front of volatile characters */
     tb_init(dst, 0);
@@ -1117,21 +1289,27 @@ kbd_show_response(TBUFF **dst,	/* string with escapes */
     }
     if (vl_echo)
 	kbd_flush();
-    return tb_length(*dst);
+    return (int) tb_length(*dst);
 }
 
 /* check only for eol-character */
 int
 /*ARGSUSED*/
-eol_null(const char *buffer GCC_UNUSED, unsigned cpos GCC_UNUSED, int c, int eolchar)
+eol_null(EOL_ARGS)
 {
+    (void) buffer;
+    (void) cpos;
+
     return (c == eolchar);
 }
 
 int
 /*ARGSUSED*/
-eol_history(const char *buffer GCC_UNUSED, unsigned cpos GCC_UNUSED, int c, int eolchar)
+eol_history(EOL_ARGS)
 {
+    (void) buffer;
+    (void) cpos;
+
     if (isPrint(eolchar) || eolchar == '\r') {
 	if (c == eolchar || (eolchar == '\r' && c == '\n'))
 	    return TRUE;
@@ -1180,7 +1358,7 @@ kbd_is_pushed_back(void)
 int
 kbd_string(const char *prompt,	/* put this out first */
 	   char *extbuf,	/* the caller's (possibly full) buffer */
-	   unsigned bufn,	/* the length of  " */
+	   size_t bufn,		/* the length of  " */
 	   int eolchar,		/* char we can terminate on, in addition to '\n' */
 	   KBD_OPTIONS options,	/* KBD_EXPAND/KBD_QUOTES, etc. */
 	   int (*complete) (DONE_ARGS))		/* handles completion */
@@ -1321,7 +1499,7 @@ fakeKeyCode(const CMDFUNC * f)
 }
 
 static int
-editMinibuffer(TBUFF **buf, unsigned *cpos, int c, int margin, int quoted)
+editMiniBuffer(TBUFF **buf, size_t *cpos, int c, int margin, int quoted)
 {
     int edited = FALSE;
     const CMDFUNC *cfp;
@@ -1343,124 +1521,137 @@ editMinibuffer(TBUFF **buf, unsigned *cpos, int c, int margin, int quoted)
 	   *cpos, tb_visible(*buf),
 	   c, miniedit, DOT.o));
 
-    cfp = ((miniedit)
-	   ? CommandKeyBinding(c)
-	   : InsertKeyBinding(c));
+#if OPT_MULTIBYTE
+    if (char2int(c) >= (int) iBIT(MinCBits)) {
+	UCHAR temp[10];
+	int used = vl_conv_to_utf8(temp, c, sizeof(temp));
+	int n;
 
-    /* Use editc (normally ^G) to toggle insert/command mode */
-    if (c == editc && !quoted) {
-	miniedit = !miniedit;
-    } else if ((miniedit || isSpecial(c))
-	       && (cfp != 0) && (cfp->c_flags & MINIBUF) != 0) {
-	int first = *cpos + margin;
-	int old_clexec = clexec;
-	int old_named = isnamedcmd;
-	int old_margin = b_left_margin(bminip);
-	REGIONSHAPE old_shape = regionshape;
+	for (n = 0; n < used; ++n) {
+	    edited |= editMiniBuffer(buf, cpos, temp[n], margin, quoted);
+	}
+    } else
+#endif
+    {
+	cfp = ((miniedit)
+	       ? CommandKeyBinding(c)
+	       : InsertKeyBinding(c));
 
-	/*
-	 * Reset flags that might cause a recursion into the prompt/reply
-	 * code.
-	 */
-	clexec = 0;
-	isnamedcmd = 0;
+	/* Use editc (normally ^G) to toggle insert/command mode */
+	if (c == editc && !quoted) {
+	    miniedit = !miniedit;
+	} else if ((miniedit || isSpecial(c))
+		   && (cfp != 0) && (cfp->c_flags & MINIBUF) != 0) {
+	    C_NUM first = (C_NUM) * cpos + margin;
+	    int old_clexec = clexec;
+	    int old_named = isnamedcmd;
+	    int old_margin = b_left_margin(bminip);
+	    REGIONSHAPE old_shape = regionshape;
 
-	/*
-	 * Set limits so we don't edit the prompt, and allow us to move the
-	 * cursor with the arrow keys just past the end of line.
-	 */
-	b_set_left_margin(bminip, margin);
-	DOT.o = llength(DOT.l);
-	linsert(1, ' ');	/* pad the line so we can move */
+	    /*
+	     * Reset flags that might cause a recursion into the prompt/reply
+	     * code.
+	     */
+	    clexec = 0;
+	    isnamedcmd = 0;
 
-	DOT.o = first;
-	MK = DOT;
-	curwp->w_line = DOT;
-	regionshape = EXACT;	/* operdel(), etc., do not set this */
-	(void) execute(cfp, FALSE, 1);
-	insertmode = save_insertmode;
-	edited = TRUE;
-
-	llength(DOT.l) -= 1;	/* strip the padding */
-	b_set_left_margin(bminip, old_margin);
-
-	clexec = old_clexec;
-	isnamedcmd = old_named;
-	regionshape = old_shape;
-
-	/*
-	 * Cheat a little, since we may have used an alias for
-	 * '$', which can set the offset just past the end of
-	 * line.
-	 */
-	if (DOT.o > llength(DOT.l)) {
+	    /*
+	     * Set limits so we don't edit the prompt, and allow us to move the
+	     * cursor with the arrow keys just past the end of line.
+	     */
+	    b_set_left_margin(bminip, margin);
 	    DOT.o = llength(DOT.l);
-	}
-	*cpos = DOT.o - margin;
+	    lins_bytes(1, ' ');	/* pad the line so we can move */
 
-	/*
-	 * Copy the data back from the minibuffer into our working TBUFF.
-	 */
-	tb_init(buf, EOS);
-	if (llength(DOT.l) > margin)
-	    tb_bappend(buf, lvalue(DOT.l) + margin, llength(DOT.l) - margin);
+	    DOT.o = first;
+	    MK = DOT;
+	    curwp->w_line = DOT;
+	    regionshape = EXACT;	/* operdel(), etc., do not set this */
+	    (void) execute(cfp, FALSE, 1);
+	    insertmode = save_insertmode;
+	    edited = TRUE;
 
-	/*
-	 * Below are some workarounds for making it appear that we're doing the
-	 * right thing for certain non-motion commands.  They allow us to revert to
-	 * insert-mode (the default), moving the cursor first as expected.
-	 */
-    } else if (cfp == &f_insert) {
-	miniedit = !miniedit;
-    } else if (miniedit && cfp == &f_insertbol) {
-	edited = editMinibuffer(buf, cpos, fakeKeyCode(&f_firstnonwhite),
-				margin, quoted);
-	miniedit = FALSE;
-    } else if (miniedit && cfp == &f_appendeol) {
-	edited = editMinibuffer(buf, cpos, fakeKeyCode(&f_gotoeol),
-				margin, quoted);
-	miniedit = FALSE;
-    } else if (miniedit && cfp == &f_append) {
-	edited = editMinibuffer(buf, cpos, fakeKeyCode(&f_forwchar_to_eol),
-				margin, quoted);
-	miniedit = FALSE;
-    } else if (isSpecial(c) || (miniedit && (cfp != 0))) {
-	/*
-	 * Reject other non-motion commands for now.  We haven't a good way to
-	 * update the minibuffer if someone inserts a newline, so we couldn't
-	 * implement insert except as a special case (see above).
-	 */
-	kbd_alarm();
-    } else {
-	/*
-	 * If it is not a command, drop out of miniedit mode and append the
-	 * character to the buffer.
-	 */
-	LINE *lp = DOT.l;
-	miniedit = FALSE;
-	if (!vl_echo || qpasswd) {
-	    char tmp = (char) c;
-	    tb_bappend(buf, &tmp, 1);
-	    if (qpasswd)
-		show1Char(c);
-	} else if (llength(lp) >= margin) {
-	    show1Char(c);
+	    llength(DOT.l) -= 1;	/* strip the padding */
+	    b_set_left_margin(bminip, old_margin);
+
+	    clexec = old_clexec;
+	    isnamedcmd = old_named;
+	    regionshape = old_shape;
+
+	    /*
+	     * Cheat a little, since we may have used an alias for
+	     * '$', which can set the offset just past the end of
+	     * line.
+	     */
+	    if (DOT.o > llength(DOT.l)) {
+		DOT.o = llength(DOT.l);
+	    }
+	    *cpos = DOT.o - margin;
+
+	    /*
+	     * Copy the data back from the minibuffer into our working TBUFF.
+	     */
 	    tb_init(buf, EOS);
-	    tb_bappend(buf, lvalue(lp) + margin, llength(lp) - margin);
+	    if (llength(DOT.l) > margin)
+		tb_bappend(buf, lvalue(DOT.l) + margin, llength(DOT.l) - margin);
+
+	    /*
+	     * Below are some workarounds for making it appear that we're doing the
+	     * right thing for certain non-motion commands.  They allow us to revert to
+	     * insert-mode (the default), moving the cursor first as expected.
+	     */
+	} else if (cfp == &f_insert) {
+	    miniedit = !miniedit;
+	} else if (miniedit && cfp == &f_insertbol) {
+	    edited = editMiniBuffer(buf, cpos, fakeKeyCode(&f_firstnonwhite),
+				    margin, quoted);
+	    miniedit = FALSE;
+	} else if (miniedit && cfp == &f_appendeol) {
+	    edited = editMiniBuffer(buf, cpos, fakeKeyCode(&f_gotoeol),
+				    margin, quoted);
+	    miniedit = FALSE;
+	} else if (miniedit && cfp == &f_append) {
+	    edited = editMiniBuffer(buf, cpos, fakeKeyCode(&f_forwchar_to_eol),
+				    margin, quoted);
+	    miniedit = FALSE;
+	} else if (isSpecial(c) || (miniedit && (cfp != 0))) {
+	    /*
+	     * Reject other non-motion commands for now.  We haven't a good way to
+	     * update the minibuffer if someone inserts a newline, so we couldn't
+	     * implement insert except as a special case (see above).
+	     */
+	    kbd_alarm();
+	} else {
+	    /*
+	     * If it is not a command, drop out of miniedit mode and append the
+	     * character to the buffer.
+	     */
+	    LINE *lp = DOT.l;
+	    miniedit = FALSE;
+	    if (!vl_echo || qpasswd) {
+		char tmp = (char) c;
+		tb_bappend(buf, &tmp, 1);
+		if (qpasswd)
+		    show1Char(c);
+	    } else if (llength(lp) >= margin) {
+		show1Char(c);
+		tb_init(buf, EOS);
+		tb_bappend(buf, lvalue(lp) + margin, llength(lp) - margin);
+	    }
+	    *cpos += 1;
+	    edited = TRUE;
 	}
-	*cpos += 1;
-	edited = TRUE;
+
+	shiftMiniBuffer(DOT.o);
+
+	if (*cpos > tb_length(*buf))
+	    *cpos = tb_length(*buf);
+
+	TRACE(("...editMiniBuffer(%d:%s) returns %d, miniedit=%d, dot=%d\n",
+	       *cpos, tb_visible(*buf),
+	       edited, miniedit, DOT.o));
     }
 
-    shiftMiniBuffer(DOT.o);
-
-    if (*cpos > tb_length(*buf))
-	*cpos = tb_length(*buf);
-
-    TRACE(("...editMiniBuffer(%d:%s) returns %d, miniedit=%d, dot=%d\n",
-	   *cpos, tb_visible(*buf),
-	   edited, miniedit, DOT.o));
-
     --no_minimsgs;
     curbp = savebp;
     curwp = savewp;
@@ -1480,30 +1671,46 @@ editMinibuffer(TBUFF **buf, unsigned *cpos, int c, int margin, int quoted)
 int
 read_quoted(int count, int inscreen)
 {
-    int c, digs, base, i, num, delta;
+    int c, digs, base, i, delta;
+    unsigned value = 0;
+    unsigned limit = 0xff;
     const char *str;
 
+    TRACE((T_CALLED "read_quoted(%d,%d)\n", count, inscreen));
+
     i = digs = 0;
-    num = 0;
 
     c = keystroke_raw8();
     if (count <= 0)
-	return c;
+	returnCode(c);
 
-    /* accumulate up to 3 digits */
-    if (isDigit(c) || c == 'x') {
+    /* accumulate up to 3 digits for a single byte */
+    if (isDigit(c)
+#if OPT_MULTIBYTE
+	|| (c == 'u')
+#endif
+	|| (c == 'x')) {
 	if (!inscreen) {
 	    kbd_putc(c);
 	    kbd_flush();
 	}
+#if OPT_MULTIBYTE
+	if (c == 'u') {
+	    c = '0';
+	    digs = 5;		/* 4 plus the fake '0' */
+	    base = 16;
+	    str = "unicode";
+	    limit = 0xffff;
+	} else
+#endif
 	if (c == '0') {
 	    digs = 4;		/* including the leading '0' */
 	    base = 8;
 	    str = "octal";
 	} else if (c == 'x') {
+	    c = '0';
 	    digs = 3;		/* including the leading 'x' */
 	    base = 16;
-	    c = '0';
 	    str = "hex";
 	} else {
 	    digs = 3;
@@ -1512,7 +1719,7 @@ read_quoted(int count, int inscreen)
 	}
 	do {
 	    if (isbackspace(c)) {
-		num /= base;
+		value /= base;
 		if (--i < 0)
 		    break;
 	    } else {
@@ -1525,12 +1732,12 @@ read_quoted(int count, int inscreen)
 		    delta = '0';
 		else
 		    break;
-		num = num * base + c - delta;
+		value = value * base + c - delta;
 		i++;
 	    }
 
 	    if (inscreen) {
-		mlwrite("Enter %s digits... %d", str, num);
+		mlwrite("Enter %s digits... %d", str, value);
 	    } else if (i > 1) {
 		kbd_putc(c);
 		kbd_flush();
@@ -1538,7 +1745,6 @@ read_quoted(int count, int inscreen)
 
 	    if (i >= digs)
 		break;
-	    (void) update(FALSE);
 	    c = keystroke_raw8();
 	} while (isbackspace(c) ||
 		 (isDigit(c) && base >= 10) ||
@@ -1557,12 +1763,12 @@ read_quoted(int count, int inscreen)
 
     if (c >= 0) {
 	if (i == 0)		/* Did we start a number? */
-	    return c;
+	    returnCode(c);
 	else if (i < digs)	/* any other character will be pushed back */
 	    unkeystroke(c);
     }
 
-    return (i > 0) ? (num & 0xff) : -1;
+    returnCode((i > 0) ? ((int) (value & limit)) : -1);
 }
 
 static int
@@ -1611,7 +1817,7 @@ kbd_reply(const char *prompt,	/* put this out first */
 {
     int c;
     int done;
-    unsigned cpos;		/* current position */
+    size_t cpos;		/* current position */
     int status;
     int shell;
     int margin;
@@ -1619,13 +1825,15 @@ kbd_reply(const char *prompt,	/* put this out first */
     MARK saveMK;		/* FIXME: may be lost by name-completion swbuffer */
 
     int quotef;			/* are we quoting the next char? */
-    UINT backslashes;		/* are we quoting the next expandable char? */
+    size_t backslashes;		/* are we quoting the next expandable char? */
     UINT dontmap = (options & KBD_NOMAP);
     int firstch = TRUE;
     int lastch;
-    unsigned newpos;
+    size_t newpos;
     TBUFF *buf = 0;
     const char *result;
+    const char *old_bname = curbp->b_bname;
+    WINDOW *old_wp = curwp;
 
     TRACE((T_CALLED "kbd_reply(prompt=%s, extbuf=%s, options=%#x)\n",
 	   TRACE_NULL(prompt),
@@ -1884,7 +2092,7 @@ kbd_reply(const char *prompt,	/* put this out first */
 					     tbreserve(&buf), &cpos));
 		    }
 		    firstch = FALSE;
-		    shiftMiniBuffer(cpos + margin);
+		    shiftMiniBuffer((int) (cpos + margin));
 		    continue;
 		}
 	    }
@@ -2002,13 +2210,13 @@ kbd_reply(const char *prompt,	/* put this out first */
 		    c = toUpper(c);
 	    }
 #endif
-	    if (!editMinibuffer(&buf, &cpos, c, margin, EscOrQuo))
+	    if (!editMiniBuffer(&buf, &cpos, c, margin, EscOrQuo))
 		continue;	/* keep firstch==TRUE */
 	}
 	firstch = FALSE;
     }
 #if OPT_POPUPCHOICE
-    popdown_completions();
+    popdown_completions(old_bname, old_wp);
 #endif
     miniedit = FALSE;
     reading_msg_line = FALSE;
@@ -2153,7 +2361,7 @@ void
 get_kbd_macro(TBUFF **rp)
 {
     char temp[80];
-    unsigned n, last, len;
+    size_t n, last, len;
 
     *rp = tb_init(rp, EOS);
     if ((last = itb_length(KbdMacro)) != 0) {
@@ -2216,7 +2424,7 @@ int
 kbd_mac_save(int f GCC_UNUSED, int n GCC_UNUSED)
 {
     TBUFF *p = 0;
-    unsigned j, len;
+    size_t j, len;
 
     ksetup();
     get_kbd_macro(&p);
